@@ -1,9 +1,9 @@
 # coding=utf-8
 #
 # This file is part of Hypothesis, which may be found at
-# https://github.com/HypothesisWorks/hypothesis-python
+# https://github.com/HypothesisWorks/hypothesis/
 #
-# Most of this work is copyright (C) 2013-2018 David R. MacIver
+# Most of this work is copyright (C) 2013-2019 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
 # CONTRIBUTING.rst for a full list of people who may hold copyright, and
 # consult the git log if you need to determine who owns an individual
@@ -11,7 +11,7 @@
 #
 # This Source Code Form is subject to the terms of the Mozilla Public License,
 # v. 2.0. If a copy of the MPL was not distributed with this file, You can
-# obtain one at http://mozilla.org/MPL/2.0/.
+# obtain one at https://mozilla.org/MPL/2.0/.
 #
 # END HEADER
 
@@ -21,12 +21,12 @@ import math
 
 import numpy as np
 
+import hypothesis._strategies as st
 import hypothesis.internal.conjecture.utils as cu
-import hypothesis.strategies as st
 from hypothesis import Verbosity
 from hypothesis._settings import note_deprecation
 from hypothesis.errors import InvalidArgument
-from hypothesis.internal.compat import hrange, text_type
+from hypothesis.internal.compat import hrange, integer_types, text_type
 from hypothesis.internal.coverage import check_function
 from hypothesis.internal.reflection import proxies
 from hypothesis.internal.validation import check_type
@@ -126,13 +126,9 @@ class ArrayStrategy(SearchStrategy):
     def __init__(self, element_strategy, shape, dtype, fill, unique):
         self.shape = tuple(shape)
         self.fill = fill
+        assert shape, "Zero-dimensional array shape is special-cased in arrays()"
         check_argument(
-            shape,
-            "Array shape must have at least one dimension, provided shape was {}",
-            shape,
-        )
-        check_argument(
-            all(isinstance(s, int) for s in shape),
+            all(isinstance(s, integer_types) for s in shape),
             "Array shape must be integer in each dimension, provided shape was {}",
             shape,
         )
@@ -175,7 +171,8 @@ class ArrayStrategy(SearchStrategy):
                 "Generated array element %r from %r cannot be represented as "
                 "dtype %r - instead it becomes %r .  Consider using a more "
                 "precise strategy, as this will be an error in a future "
-                "version." % (val, strategy, self.dtype, result[idx])
+                "version." % (val, strategy, self.dtype, result[idx]),
+                since="2018-10-25",
             )
             # Because the message includes the value of the generated element,
             # it would be easy to spam users with thousands of warnings.
@@ -392,8 +389,10 @@ def arrays(
         shape = (shape,)
     shape = tuple(shape)
     if not shape:
-        if dtype.kind != u"O":
-            return draw(elements)
+        # We use .itemset() instead of np.full because the latter cannot set tuples as elements.
+        arr = np.empty(shape=(), dtype=dtype)
+        arr.itemset(draw(elements))
+        return arr
     fill = fill_for(elements=elements, unique=unique, fill=fill)
     return draw(ArrayStrategy(elements, shape, dtype, fill, unique))
 
@@ -402,8 +401,8 @@ def arrays(
 def array_shapes(min_dims=1, max_dims=3, min_side=1, max_side=10):
     # type: (int, int, int, int) -> st.SearchStrategy[Tuple[int, ...]]
     """Return a strategy for array shapes (tuples of int >= 1)."""
-    order_check("dims", 1, min_dims, max_dims)
-    order_check("side", 1, min_side, max_side)
+    order_check("dims", 0, min_dims, max_dims)
+    order_check("side", 0, min_side, max_side)
     return st.lists(
         st.integers(min_side, max_side), min_size=min_dims, max_size=max_dims
     ).map(tuple)
@@ -429,7 +428,20 @@ def defines_dtype_strategy(strat):
     @st.defines_strategy
     @proxies(strat)
     def inner(*args, **kwargs):
-        return strat(*args, **kwargs).map(np.dtype)
+        strategy = strat(*args, **kwargs)
+
+        def convert_to_dtype(x):
+            """Helper to debug issue #1798."""
+            try:
+                return np.dtype(x)
+            except ValueError:
+                print(
+                    "Got invalid dtype value=%r from strategy=%r, function=%r"
+                    % (x, strategy, strat)
+                )
+                raise
+
+        return strategy.map(convert_to_dtype)
 
     return inner
 
