@@ -33,11 +33,25 @@ from hypothesis.internal.compat import (
 )
 from hypothesis.internal.conjecture.utils import calc_label_from_name
 from hypothesis.internal.escalation import mark_for_escalation
+from hypothesis.utils.conventions import UniqueIdentifier
 
 TOP_LABEL = calc_label_from_name("top")
 DRAW_BYTES_LABEL = calc_label_from_name("draw_bytes() in ConjectureData")
 
 
+class ExtraInformation(object):
+    """A class for holding shared state on a ``ConjectureData`` that should
+    be added to the final ``ConjectureResult``."""
+
+    def __repr__(self):
+        return "ExtraInformation(%s)" % (
+            ", ".join(["%s=%r" % (k, v) for k, v in self.__dict__.items()]),
+        )
+
+    def has_information(self):
+        return bool(self.__dict__)
+
+
 class Status(IntEnum):
     OVERRUN = 0
     INVALID = 1
@@ -68,7 +82,7 @@ class Example(object):
 
     # Depth of this example in the example tree. The top-level example has a
     # depth of 0.
-    depth = attr.ib()
+    depth = attr.ib(repr=False)
 
     # A label is an opaque value that associates each example with its
     # approximate origin, such as a particular strategy class or a particular
@@ -84,16 +98,16 @@ class Example(object):
     # An example is "trivial" if it only contains forced bytes and zero bytes.
     # All examples start out as trivial, and then get marked non-trivial when
     # we see a byte that is neither forced nor zero.
-    trivial = attr.ib(default=True)
+    trivial = attr.ib(default=True, repr=False)
 
     # True if we believe that the shrinker should be able to delete this
     # example completely, without affecting the value produced by its enclosing
     # strategy. Typically set when a rejection sampler decides to reject a
     # generated value and try again.
-    discarded = attr.ib(default=None)
+    discarded = attr.ib(default=None, repr=False)
 
     # List of child examples, represented as indices into the example list.
-    children = attr.ib(default=attr.Factory(list))
+    children = attr.ib(default=attr.Factory(list), repr=False)
 
     @property
     def length(self):
@@ -119,11 +133,11 @@ class Block(object):
     # True if this block's byte values were forced by a write operation.
     # As long as the bytes before this block remain the same, modifying this
     # block's bytes will have no effect.
-    forced = attr.ib()
+    forced = attr.ib(repr=False)
 
     # True if this block's byte values are all 0. Reading this flag can be
     # more convenient than explicitly checking a slice for non-zero bytes.
-    all_zero = attr.ib()
+    all_zero = attr.ib(repr=False)
 
     @property
     def bounds(self):
@@ -153,6 +167,97 @@ global_test_counter = 0
 MAX_DEPTH = 100
 
 
+def calc_examples(self):
+    """Build the list of examples from either a ``ConjectureResult``
+    or a ``ConjectureData`` object by interpreting the recorded
+    example boundaries and parsing them into a tree of ``Example``
+    objects, returning the resulting list.
+
+    This is needed because we want to calculate these lazily.
+    The ``Example`` tree is fairly memory hungry and mildly
+    expensive to compute and, especially during generation, we
+    will often not need it, so we only want to compute it on
+    demand."""
+    assert self.example_boundaries
+
+    example_stack = []
+    examples = []
+
+    non_trivial_block_starts = {
+        b.start for b in self.blocks if not (b.all_zero or b.forced)
+    }
+
+    for index, labels in self.example_boundaries:
+        for label in labels:
+            if label in (Stop, StopDiscard):
+                discard = label is StopDiscard
+                k = example_stack.pop()
+                ex = examples[k]
+                ex.end = index
+
+                if ex.length == 0:
+                    ex.trivial = True
+
+                if example_stack and not ex.trivial:
+                    examples[example_stack[-1]].trivial = False
+
+                ex.discarded = discard
+            else:
+                i = len(examples)
+                ex = Example(
+                    index=i,
+                    depth=len(example_stack),
+                    label=label,
+                    start=index,
+                    trivial=index not in non_trivial_block_starts,
+                )
+                examples.append(ex)
+                if example_stack:
+                    p = example_stack[-1]
+                    examples[p].children.append(ex)
+                example_stack.append(i)
+    for ex in examples:
+        assert ex.end is not None
+
+    assert examples
+    return examples
+
+
+@attr.s(slots=True)
+class ConjectureResult(object):
+    """Result class storing the parts of ConjectureData that we
+    will care about after the original ConjectureData has outlived its
+    usefulness."""
+
+    status = attr.ib()
+    interesting_origin = attr.ib()
+    buffer = attr.ib()
+    blocks = attr.ib()
+    example_boundaries = attr.ib()
+    output = attr.ib()
+    extra_information = attr.ib()
+    __examples = attr.ib(init=False, default=None)
+
+    index = attr.ib(init=False)
+
+    def __attrs_post_init__(self):
+        self.index = len(self.buffer)
+
+    @property
+    def examples(self):
+        if self.__examples is None:
+            self.__examples = calc_examples(self)
+            self.example_boundaries = None
+
+        assert self.example_boundaries is None
+        return self.__examples
+
+
+# Special "labels" used to indicate the end of example boundaries
+Stop = UniqueIdentifier("Stop")
+StopDiscard = UniqueIdentifier("StopDiscard")
+
+
 class ConjectureData(object):
     @classmethod
     def for_buffer(self, buffer):
@@ -170,6 +275,7 @@ class ConjectureData(object):
         self.block_starts = {}
         self.blocks = []
         self.buffer = bytearray()
+        self.index = 0
         self.output = u""
         self.status = Status.VALID
         self.frozen = False
@@ -184,12 +290,21 @@ class ConjectureData(object):
         self.draw_times = []
         self.max_depth = 0
 
-        self.examples = []
-        self.example_stack = []
-        self.has_discards = False
+        self.example_boundaries = []
 
-        top = self.start_example(TOP_LABEL)
-        assert top.depth == 0
+        self.__result = None
+
+        # Normally unpopulated but we need this in the niche case
+        # that self.as_result() is Overrun but we still want the
+        # examples for reporting purposes.
+        self.__examples = None
+
+        # We want the top level example to have depth 0, so we start
+        # at -1.
+        self.depth = -1
+
+        self.start_example(TOP_LABEL)
+        self.extra_information = ExtraInformation()
 
     def __repr__(self):
         return "ConjectureData(%s, %d bytes%s)" % (
@@ -198,23 +313,31 @@ class ConjectureData(object):
             ", frozen" if self.frozen else "",
         )
 
+    def as_result(self):
+        """Convert the result of running this test into
+        either an Overrun object or a ConjectureResult."""
+
+        assert self.frozen
+        if self.status == Status.OVERRUN:
+            return Overrun
+        if self.__result is None:
+            self.__result = ConjectureResult(
+                status=self.status,
+                interesting_origin=self.interesting_origin,
+                buffer=self.buffer,
+                example_boundaries=self.example_boundaries,
+                blocks=self.blocks,
+                output=self.output,
+                extra_information=self.extra_information
+                if self.extra_information.has_information()
+                else None,
+            )
+        return self.__result
+
     def __assert_not_frozen(self, name):
         if self.frozen:
             raise Frozen("Cannot call %s on frozen ConjectureData" % (name,))
 
-    @property
-    def depth(self):
-        # We always have a single example wrapping everything. We want to treat
-        # that as depth 0 rather than depth 1.
-        return len(self.example_stack) - 1
-
-    @property
-    def index(self):
-        return len(self.buffer)
-
-    def all_block_bounds(self):
-        return [block.bounds for block in self.blocks]
-
     def note(self, value):
         self.__assert_not_frozen("note")
         if not isinstance(value, text_type):
@@ -261,71 +384,51 @@ class ConjectureData(object):
         finally:
             self.stop_example()
 
+    def current_example_labels(self):
+        if not self.example_boundaries or self.example_boundaries[-1][0] < self.index:
+            self.example_boundaries.append((self.index, []))
+        return self.example_boundaries[-1][-1]
+
     def start_example(self, label):
         self.__assert_not_frozen("start_example")
-
-        i = len(self.examples)
-        new_depth = self.depth + 1
-        ex = Example(index=i, depth=new_depth, label=label, start=self.index)
-        self.examples.append(ex)
-        if self.example_stack:
-            p = self.example_stack[-1]
-            self.examples[p].children.append(ex)
-        self.example_stack.append(i)
+        self.current_example_labels().append(label)
+        self.depth += 1
         self.max_depth = max(self.max_depth, self.depth)
-        return ex
 
     def stop_example(self, discard=False):
         if self.frozen:
             return
-
-        k = self.example_stack.pop()
-        ex = self.examples[k]
-        ex.end = self.index
-
-        if self.example_stack and not ex.trivial:
-            self.examples[self.example_stack[-1]].trivial = False
-
-        # We don't want to count empty examples as discards even if the flag
-        # says we should. This leads to situations like
-        # https://github.com/HypothesisWorks/hypothesis/issues/1230
-        # where it can look like we should discard data but there's nothing
-        # useful for us to do.
-        if self.index == ex.start:
-            discard = False
-
-        ex.discarded = discard
-
-        if discard:
-            self.has_discards = True
+        self.current_example_labels().append(StopDiscard if discard else Stop)
+        self.depth -= 1
+        assert self.depth >= -1
 
     def note_event(self, event):
         self.events.add(event)
 
+    @property
+    def examples(self):
+        result = self.as_result()
+        if result is Overrun:
+            if self.__examples is None:
+                self.__examples = calc_examples(self)
+            return self.__examples
+        else:
+            return result.examples
+
     def freeze(self):
         if self.frozen:
             assert isinstance(self.buffer, hbytes)
             return
         self.finish_time = benchmark_time()
+        assert len(self.buffer) == self.index
 
-        while self.example_stack:
+        # Always finish by closing all remaining examples so that we have a
+        # valid tree.
+        while self.depth >= 0:
             self.stop_example()
 
         self.frozen = True
 
-        if self.status >= Status.VALID:
-            discards = []
-            for ex in self.examples:
-                if ex.length == 0:
-                    continue
-                if discards:
-                    u, v = discards[-1]
-                    if u <= ex.start <= ex.end <= v:
-                        continue
-                if ex.discarded:
-                    discards.append((ex.start, ex.end))
-                    continue
-
         self.buffer = hbytes(self.buffer)
         self.events = frozenset(self.events)
         del self._draw_bytes
@@ -368,7 +471,7 @@ class ConjectureData(object):
             raise StopTest(self.testcounter)
 
     def __write(self, result, forced=False):
-        ex = self.start_example(DRAW_BYTES_LABEL)
+        self.start_example(DRAW_BYTES_LABEL)
         initial = self.index
         n = len(result)
 
@@ -379,7 +482,6 @@ class ConjectureData(object):
             forced=forced,
             all_zero=not any(result),
         )
-        ex.trivial = block.trivial
 
         self.block_starts.setdefault(n, []).append(block.start)
         self.blocks.append(block)
@@ -387,6 +489,7 @@ class ConjectureData(object):
         assert len(result) == n
         assert self.index == initial
         self.buffer.extend(result)
+        self.index = len(self.buffer)
         self.stop_example()
 
     def draw_bytes(self, n):
